@@ -7,12 +7,14 @@ from .cip import *
 from .coinselection import *
 from .crypto import *
 from .exception import *
+from .governance import *
 from .hash import *
 from .key import *
 from .metadata import *
 from .nativescript import *
 from .network import *
 from .plutus import *
+from .pool_params import *
 from .serialization import *
 from .transaction import *
 from .txbuilder import *
