@@ -3,7 +3,6 @@
 from __future__ import annotations
 
 import json
-import os
 from typing import Optional, Type
 
 from nacl.encoding import RawEncoder
@@ -32,6 +31,9 @@ __all__ = [
     "StakeSigningKey",
     "StakeVerificationKey",
     "StakeKeyPair",
+    "StakePoolSigningKey",
+    "StakePoolVerificationKey",
+    "StakePoolKeyPair",
 ]
 
 
@@ -71,7 +73,7 @@ class Key(CBORSerializable):
     def from_primitive(cls: Type["Key"], value: bytes) -> Key:
         return cls(value)
 
-    def to_json(self) -> str:
+    def to_json(self, **kwargs) -> str:  # type: ignore
         """Serialize the key to JSON.
 
         The json output has three fields: "type", "description", and "cborHex".
@@ -83,12 +85,12 @@ class Key(CBORSerializable):
             {
                 "type": self.key_type,
                 "description": self.description,
-                "cborHex": self.to_cbor(),
+                "cborHex": self.to_cbor_hex(),
             }
         )
 
     @classmethod
-    def from_json(cls: Type[Key], data: str, validate_type=False) -> Key:
+    def from_json(cls: Type[Key], data: str, validate_type=False) -> Key:  # type: ignore
         """Restore a key from a JSON string.
 
         Args:
@@ -120,18 +122,6 @@ class Key(CBORSerializable):
             description=obj["description"],
         )
 
-    def save(self, path: str):
-        if os.path.isfile(path):
-            if os.stat(path).st_size > 0:
-                raise IOError(f"File {path} already exists!")
-        with open(path, "w") as f:
-            f.write(self.to_json())
-
-    @classmethod
-    def load(cls, path: str):
-        with open(path) as f:
-            return cls.from_json(f.read())
-
     def __bytes__(self):
         return self.payload
 
@@ -208,8 +198,8 @@ class ExtendedSigningKey(Key):
 
         return cls(
             payload=hdwallet.xprivate_key + hdwallet.public_key + hdwallet.chain_code,
-            key_type="PaymentExtendedSigningKeyShelley_ed25519_bip32",
-            description="Payment Signing Key",
+            key_type=cls.KEY_TYPE,
+            description=cls.DESCRIPTION,
         )
 
 
@@ -314,3 +304,44 @@ class StakeKeyPair:
         cls: Type[StakeKeyPair], signing_key: SigningKey
     ) -> StakeKeyPair:
         return cls(signing_key, StakeVerificationKey.from_signing_key(signing_key))
+
+    def __eq__(self, other):
+        if isinstance(other, StakeKeyPair):
+            return (
+                other.signing_key == self.signing_key
+                and other.verification_key == self.verification_key
+            )
+
+
+class StakePoolSigningKey(SigningKey):
+    KEY_TYPE = "StakePoolSigningKey_ed25519"
+    DESCRIPTION = "Stake Pool Operator Signing Key"
+
+
+class StakePoolVerificationKey(VerificationKey):
+    KEY_TYPE = "StakePoolVerificationKey_ed25519"
+    DESCRIPTION = "Stake Pool Operator Verification Key"
+
+
+class StakePoolKeyPair:
+    def __init__(self, signing_key: SigningKey, verification_key: VerificationKey):
+        self.signing_key = signing_key
+        self.verification_key = verification_key
+
+    @classmethod
+    def generate(cls: Type[StakePoolKeyPair]) -> StakePoolKeyPair:
+        signing_key = StakePoolSigningKey.generate()
+        return cls.from_signing_key(signing_key)
+
+    @classmethod
+    def from_signing_key(
+        cls: Type[StakePoolKeyPair], signing_key: SigningKey
+    ) -> StakePoolKeyPair:
+        return cls(signing_key, StakePoolVerificationKey.from_signing_key(signing_key))
+
+    def __eq__(self, other):
+        if isinstance(other, StakePoolKeyPair):
+            return (
+                other.signing_key == self.signing_key
+                and other.verification_key == self.verification_key
+            )
