@@ -10,6 +10,7 @@ from nacl.hash import blake2b
 from pycardano.exception import DeserializeException, InvalidArgumentException
 from pycardano.hash import AUXILIARY_DATA_HASH_SIZE, AuxiliaryDataHash
 from pycardano.nativescript import NativeScript
+from pycardano.plutus import PlutusV1Script, PlutusV2Script, PlutusV3Script
 from pycardano.serialization import (
     ArrayCBORSerializable,
     CBORSerializable,
@@ -41,6 +42,7 @@ class Metadata(DictCBORSerializable):
                 if len(data) > self.MAX_ITEM_SIZE:
                     raise InvalidArgumentException(
                         f"The size of {data} exceeds {self.MAX_ITEM_SIZE} bytes."
+                        "Use pycardano.serialization.ByteString for long bytes."
                     )
             elif isinstance(data, str):
                 if len(data.encode("utf-8")) > self.MAX_ITEM_SIZE:
@@ -86,9 +88,15 @@ class AlonzoMetadata(MapCBORSerializable):
         default=None,
         metadata={"optional": True, "key": 1, "object_hook": list_hook(NativeScript)},
     )
-    plutus_scripts: Optional[List[bytes]] = field(
+    plutus_v1_scripts: Optional[List[PlutusV1Script]] = field(
         default=None, metadata={"optional": True, "key": 2}
     )
+    plutus_v2_scripts: Optional[List[PlutusV2Script]] = field(
+        default=None, metadata={"optional": True, "key": 3}
+    )
+    plutus_v3_scripts: Optional[List[PlutusV3Script]] = field(
+        default=None, metadata={"optional": True, "key": 4}
+    )
 
     def to_primitive(self) -> Primitive:
         return CBORTag(AlonzoMetadata.TAG, super(AlonzoMetadata, self).to_primitive())
@@ -115,7 +123,9 @@ class AuxiliaryData(CBORSerializable):
         return self.data.to_primitive()
 
     @classmethod
-    def from_primitive(cls: Type[AuxiliaryData], value: Primitive) -> AuxiliaryData:
+    def from_primitive(
+        cls: Type[AuxiliaryData], value: Primitive, type_args: Optional[tuple] = None
+    ) -> AuxiliaryData:
         for t in [AlonzoMetadata, ShelleyMarryMetadata, Metadata]:
             # The schema of metadata in different eras are mutually exclusive, so we can try deserializing
             # them one by one without worrying about mismatch.
@@ -127,5 +137,5 @@ class AuxiliaryData(CBORSerializable):
 
     def hash(self) -> AuxiliaryDataHash:
         return AuxiliaryDataHash(
-            blake2b(self.to_cbor("bytes"), AUXILIARY_DATA_HASH_SIZE, encoder=RawEncoder)  # type: ignore
+            blake2b(self.to_cbor(), AUXILIARY_DATA_HASH_SIZE, encoder=RawEncoder)  # type: ignore
         )
