@@ -35,7 +35,7 @@ class NativeScript(ArrayCBORSerializable):
     json_field: ClassVar[str]
 
     @classmethod
-    @limit_primitive_type(list)
+    @limit_primitive_type(list, tuple)
     def from_primitive(
         cls: Type[NativeScript], value: list
     ) -> Union[
@@ -58,7 +58,7 @@ class NativeScript(ArrayCBORSerializable):
             raise DeserializeException(f"Unknown script type indicator: {script_type}")
 
     def hash(self) -> ScriptHash:
-        cbor_bytes = cast(bytes, self.to_cbor("bytes"))
+        cbor_bytes = cast(bytes, self.to_cbor())
         return ScriptHash(
             blake2b(bytes(1) + cbor_bytes, SCRIPT_HASH_SIZE, encoder=RawEncoder)
         )
