@@ -3,8 +3,9 @@
 from __future__ import annotations
 
 from enum import Enum
+from typing import Type
 
-from pycardano.serialization import CBORSerializable
+from pycardano.serialization import CBORSerializable, limit_primitive_type
 
 __all__ = ["Network"]
 
@@ -21,5 +22,6 @@ class Network(CBORSerializable, Enum):
         return self.value
 
     @classmethod
-    def from_primitive(cls, value: int) -> Network:
+    @limit_primitive_type(int)
+    def from_primitive(cls: Type[Network], value: int) -> Network:
         return cls(value)
