@@ -4,20 +4,24 @@ from __future__ import annotations
 
 import inspect
 import json
+import typing
 from dataclasses import dataclass, field, fields
 from enum import Enum
-from typing import Any, ClassVar, Optional, Type, Union
+from hashlib import sha256
+from typing import Any, List, Optional, Type, Union
 
 import cbor2
 from cbor2 import CBORTag
 from nacl.encoding import RawEncoder
 from nacl.hash import blake2b
+from typeguard import typechecked
 
-from pycardano.exception import DeserializeException
+from pycardano.exception import DeserializeException, InvalidArgumentException
 from pycardano.hash import DATUM_HASH_SIZE, SCRIPT_HASH_SIZE, DatumHash, ScriptHash
 from pycardano.nativescript import NativeScript
 from pycardano.serialization import (
     ArrayCBORSerializable,
+    ByteString,
     CBORSerializable,
     DictCBORSerializable,
     IndefiniteList,
@@ -36,23 +40,36 @@ __all__ = [
     "Datum",
     "RedeemerTag",
     "ExecutionUnits",
+    "PlutusScript",
     "PlutusV1Script",
     "PlutusV2Script",
+    "PlutusV3Script",
     "RawPlutusData",
     "Redeemer",
+    "RedeemerKey",
+    "RedeemerValue",
+    "RedeemerMap",
+    "Redeemers",
     "ScriptType",
     "datum_hash",
     "plutus_script_hash",
     "script_hash",
+    "Unit",
 ]
 
 
+# taken from https://stackoverflow.com/a/13624858
+class classproperty(property):
+    def __get__(self, owner_self, owner_cls):
+        return self.fget(owner_cls)
+
+
 class CostModels(DictCBORSerializable):
     KEY_TYPE = int
     VALUE_TYPE = dict
 
     def to_shallow_primitive(self) -> dict:
-        result = {}
+        result: dict[bytes, Union[typing.List[Any], bytes]] = {}
         for language in sorted(self.keys()):
             cost_model = self[language]
             if language == 0:
@@ -64,7 +81,7 @@ class CostModels(DictCBORSerializable):
                 cm = IndefiniteList([cost_model[k] for k in sorted(cost_model.keys())])
                 result[l_cbor] = cbor2.dumps(cm, default=default_encoder)
             else:
-                result[language] = [cost_model[k] for k in sorted(cost_model.keys())]
+                result[language] = [cost_model[k] for k in cost_model.keys()]
         return result
 
     @classmethod
@@ -439,6 +456,65 @@ def get_tag(constr_id: int) -> Optional[int]:
         return None
 
 
+def get_constructor_id_and_fields(
+    raw_tag: CBORTag,
+) -> typing.Tuple[int, typing.List[Any]]:
+    tag = raw_tag.tag
+    if tag == 102:
+        if len(raw_tag.value) != 2:
+            raise DeserializeException(
+                f"Expect the length of value to be exactly 2, got {len(raw_tag.value)} instead."
+            )
+        return raw_tag.value[0], raw_tag.value[1]
+    else:
+        if 121 <= tag < 128:
+            constr = tag - 121
+        elif 1280 <= tag < 1536:
+            constr = tag - 1280 + 7
+        else:
+            raise DeserializeException(f"Unexpected tag for RawPlutusData: {tag}")
+        return constr, raw_tag.value
+
+
+def id_map(cls, skip_constructor=False):
+    """
+    Constructs a unique representation of a PlutusData type definition.
+    Intended for automatic constructor generation.
+    """
+    if cls == bytes or cls == ByteString:
+        return "bytes"
+    if cls == int:
+        return "int"
+    if cls == RawCBOR or cls == RawPlutusData or cls == Datum:
+        return "any"
+    if cls == IndefiniteList:
+        return "list"
+    if hasattr(cls, "__origin__"):
+        origin = getattr(cls, "__origin__")
+        if origin == list:
+            prefix = "list"
+        elif origin == dict:
+            prefix = "map"
+        elif origin == typing.Union:
+            prefix = "union"
+        else:
+            raise TypeError(
+                f"Unexpected parameterized type for automatic constructor generation: {cls}"
+            )
+        return prefix + "<" + ",".join(id_map(a) for a in cls.__args__) + ">"
+    if issubclass(cls, PlutusData):
+        return (
+            "cons["
+            + cls.__name__
+            + "]("
+            + (str(cls.CONSTR_ID) if not skip_constructor else "_")
+            + ";"
+            + ",".join(f.name + ":" + id_map(f.type) for f in fields(cls))
+            + ")"
+        )
+    raise TypeError(f"Unexpected type for automatic constructor generation: {cls}")
+
+
 @dataclass(repr=False)
 class PlutusData(ArrayCBORSerializable):
     """
@@ -455,23 +531,52 @@ class PlutusData(ArrayCBORSerializable):
         ...     a: int
         ...     b: bytes
         >>> test = Test(123, b"321")
-        >>> test.to_cbor()
+        >>> test.to_cbor_hex()
         'd87a9f187b43333231ff'
         >>> assert test == Test.from_cbor("d87a9f187b43333231ff")
     """
 
-    CONSTR_ID: ClassVar[int] = 0
-    """Constructor ID of this plutus data.
-       It is primarily used by Plutus core to reconstruct a data structure from serialized CBOR bytes."""
+    MAX_BYTES_SIZE = 64
+
+    @classproperty
+    def CONSTR_ID(cls):
+        """
+        Constructor ID of this plutus data.
+        It is primarily used by Plutus core to reconstruct a data structure from serialized CBOR bytes.
+        The default implementation is an almost unique, deterministic constructor ID in the range 1 - 2^32 based
+        on class attributes, types and class name.
+        """
+        k = f"_CONSTR_ID_{cls.__name__}"
+        if not hasattr(cls, k):
+            det_string = id_map(cls, skip_constructor=True)
+            det_hash = sha256(det_string.encode("utf8")).hexdigest()
+            setattr(cls, k, int(det_hash, 16) % 2**32)
+
+        return getattr(cls, k)
 
     def __post_init__(self):
-        valid_types = (PlutusData, dict, IndefiniteList, int, bytes)
+        valid_types = (
+            RawPlutusData,
+            PlutusData,
+            dict,
+            IndefiniteList,
+            int,
+            ByteString,
+            bytes,
+        )
         for f in fields(self):
             if inspect.isclass(f.type) and not issubclass(f.type, valid_types):
                 raise TypeError(
                     f"Invalid field type: {f.type}. A field in PlutusData should be one of {valid_types}"
                 )
 
+            data = getattr(self, f.name)
+            if isinstance(data, bytes) and len(data) > 64:
+                raise InvalidArgumentException(
+                    f"The size of {data} exceeds {self.MAX_BYTES_SIZE} bytes. "
+                    "Use pycardano.serialization.ByteString for long bytes."
+                )
+
     def to_shallow_primitive(self) -> CBORTag:
         primitives: Primitive = super().to_shallow_primitive()
         if primitives:
@@ -509,14 +614,12 @@ class PlutusData(ArrayCBORSerializable):
     def hash(self) -> DatumHash:
         return datum_hash(self)
 
-    def to_json(self, **kwargs) -> str:
-        """Convert to a json string
-
-        Args:
-            **kwargs: Extra key word arguments to be passed to `json.dumps()`
+    def to_dict(self) -> dict:
+        """
+        Convert to a dictionary.
 
         Returns:
-            str: a JSON encoded PlutusData.
+            str: a dict PlutusData that can be JSON encoded.
         """
 
         def _dfs(obj):
@@ -529,6 +632,8 @@ class PlutusData(ArrayCBORSerializable):
                 return {"int": obj}
             elif isinstance(obj, bytes):
                 return {"bytes": obj.hex()}
+            elif isinstance(obj, ByteString):
+                return {"bytes": obj.value.hex()}
             elif isinstance(obj, IndefiniteList) or isinstance(obj, list):
                 return {"list": [_dfs(item) for item in obj]}
             elif isinstance(obj, dict):
@@ -538,10 +643,26 @@ class PlutusData(ArrayCBORSerializable):
                     "constructor": obj.CONSTR_ID,
                     "fields": [_dfs(getattr(obj, f.name)) for f in fields(obj)],
                 }
+            elif isinstance(obj, RawPlutusData):
+                return obj.to_dict()
+            elif isinstance(obj, RawCBOR):
+                return RawPlutusData.from_cbor(obj.cbor).to_dict()
             else:
                 raise TypeError(f"Unexpected type {type(obj)}")
 
-        return json.dumps(_dfs(self), **kwargs)
+        return _dfs(self)
+
+    def to_json(self, **kwargs) -> str:  # type: ignore
+        """Convert to a json string
+
+        Args:
+            **kwargs: Extra key word arguments to be passed to `json.dumps()`
+
+        Returns:
+            str: a JSON encoded PlutusData.
+        """
+
+        return json.dumps(self.to_dict(), **kwargs)
 
     @classmethod
     def from_dict(cls: Type[PlutusData], data: dict) -> PlutusData:
@@ -559,7 +680,7 @@ class PlutusData(ArrayCBORSerializable):
                 if "constructor" in obj:
                     if obj["constructor"] != cls.CONSTR_ID:
                         raise DeserializeException(
-                            f"Mismatch between constructors, expect: {cls.CONSTR_ID}, "
+                            f"Mismatch between constructors in class {cls.__name__}, expect: {cls.CONSTR_ID}, "
                             f"got: {obj['constructor']} instead."
                         )
                     converted_fields = []
@@ -568,6 +689,8 @@ class PlutusData(ArrayCBORSerializable):
                             f_info.type, PlutusData
                         ):
                             converted_fields.append(f_info.type.from_dict(f))
+                        elif f_info.type == Datum:
+                            converted_fields.append(RawPlutusData.from_dict(f))
                         elif (
                             hasattr(f_info.type, "__origin__")
                             and f_info.type.__origin__ is Union
@@ -643,7 +766,10 @@ class PlutusData(ArrayCBORSerializable):
                 elif "int" in obj:
                     return obj["int"]
                 elif "bytes" in obj:
-                    return bytes.fromhex(obj["bytes"])
+                    if len(obj["bytes"]) > 64:
+                        return ByteString(bytes.fromhex(obj["bytes"]))
+                    else:
+                        return bytes.fromhex(obj["bytes"])
                 elif "list" in obj:
                     return IndefiniteList([_dfs(item) for item in obj["list"]])
                 else:
@@ -666,12 +792,18 @@ class PlutusData(ArrayCBORSerializable):
         obj = json.loads(data)
         return cls.from_dict(obj)
 
+    def __deepcopy__(self, memo):
+        return self.__class__.from_cbor(self.to_cbor_hex())
 
-@dataclass
+
+RawDatum = Union[PlutusData, dict, int, bytes, IndefiniteList, RawCBOR, CBORTag]
+
+
+@dataclass(repr=True)
 class RawPlutusData(CBORSerializable):
-    data: CBORTag
+    data: RawDatum
 
-    def to_primitive(self) -> CBORTag:
+    def to_primitive(self) -> Primitive:
         def _dfs(obj):
             if isinstance(obj, list) and obj:
                 return IndefiniteList([_dfs(item) for item in obj])
@@ -687,13 +819,113 @@ class RawPlutusData(CBORSerializable):
 
         return _dfs(self.data)
 
+    def to_dict(self) -> dict:
+        """
+        Convert to a dictionary.
+
+        Returns:
+            str: a dict RawPlutusData that can be JSON encoded.
+        """
+
+        def _dfs(obj):
+            if isinstance(obj, int):
+                return {"int": obj}
+            elif isinstance(obj, bytes):
+                return {"bytes": obj.hex()}
+            elif isinstance(obj, ByteString):
+                return {"bytes": obj.value.hex()}
+            elif isinstance(obj, IndefiniteList) or isinstance(obj, list):
+                return {"list": [_dfs(item) for item in obj]}
+            elif isinstance(obj, dict):
+                return {"map": [{"v": _dfs(v), "k": _dfs(k)} for k, v in obj.items()]}
+            elif isinstance(obj, CBORTag):
+                constructor, fields = get_constructor_id_and_fields(obj)
+                return {"constructor": constructor, "fields": [_dfs(f) for f in fields]}
+            elif isinstance(obj, RawCBOR):
+                return RawPlutusData.from_cbor(obj.cbor).to_dict()
+            raise TypeError(f"Unexpected type {type(obj)}")
+
+        return _dfs(RawPlutusData.to_primitive(self))
+
+    def to_json(self, **kwargs) -> str:  # type: ignore
+        """Convert to a json string
+
+        Args:
+            **kwargs: Extra key word arguments to be passed to `json.dumps()`
+
+        Returns:
+            str: a JSON encoded RawPlutusData.
+        """
+
+        return json.dumps(RawPlutusData.to_dict(self), **kwargs)
+
     @classmethod
-    @limit_primitive_type(CBORTag)
-    def from_primitive(cls: Type[RawPlutusData], value: CBORTag) -> RawPlutusData:
+    @limit_primitive_type(
+        PlutusData, dict, int, bytes, IndefiniteList, RawCBOR, CBORTag
+    )  # equal to RawDatum parameter list
+    def from_primitive(cls: Type[RawPlutusData], value: RawDatum) -> RawPlutusData:
         return cls(value)
 
+    @classmethod
+    def from_dict(cls: Type[RawPlutusData], data: dict) -> RawPlutusData:
+        """Convert a dictionary to RawPlutusData
 
-Datum = Union[PlutusData, dict, IndefiniteList, int, bytes, RawCBOR, RawPlutusData]
+        Args:
+            data (dict): A dictionary.
+
+        Returns:
+            RawPlutusData: Restored RawPlutusData.
+        """
+
+        def _dfs(obj):
+            if isinstance(obj, dict):
+                if "constructor" in obj:
+                    converted_fields = []
+                    for f in obj["fields"]:
+                        converted_fields.append(_dfs(f))
+                    tag = get_tag(obj["constructor"])
+                    if tag is None:
+                        return CBORTag(
+                            102, [obj["constructor"], IndefiniteList(converted_fields)]
+                        )
+                    else:
+                        return CBORTag(tag, converted_fields)
+                elif "map" in obj:
+                    return {_dfs(pair["k"]): _dfs(pair["v"]) for pair in obj["map"]}
+                elif "int" in obj:
+                    return obj["int"]
+                elif "bytes" in obj:
+                    if len(obj["bytes"]) > 64:
+                        return ByteString(bytes.fromhex(obj["bytes"]))
+                    else:
+                        return bytes.fromhex(obj["bytes"])
+                elif "list" in obj:
+                    return IndefiniteList([_dfs(item) for item in obj["list"]])
+                else:
+                    raise DeserializeException(f"Unexpected data structure: {obj}")
+            else:
+                raise TypeError(f"Unexpected data type: {type(obj)}")
+
+        return cls(_dfs(data))
+
+    @classmethod
+    def from_json(cls: Type[RawPlutusData], data: str) -> RawPlutusData:
+        """Restore a json encoded string to a RawPlutusData.
+
+        Args:
+            data (str): An encoded json string.
+
+        Returns:
+            RawPlutusData: The restored RawPlutusData.
+        """
+        obj = json.loads(data)
+        return cls.from_dict(obj)
+
+    def __deepcopy__(self, memo):
+        return self.__class__.from_cbor(self.to_cbor_hex())
+
+
+Datum = Union[PlutusData, dict, int, bytes, IndefiniteList, RawCBOR, RawPlutusData]
 """Plutus Datum type. A Union type that contains all valid datum types."""
 
 
@@ -714,8 +946,10 @@ class RedeemerTag(CBORSerializable, Enum):
 
     SPEND = 0
     MINT = 1
-    CERT = 2
-    REWARD = 3
+    CERTIFICATE = 2
+    WITHDRAWAL = 3
+    VOTING = 4
+    PROPOSING = 5
 
     def to_primitive(self) -> int:
         return self.value
@@ -739,6 +973,12 @@ class ExecutionUnits(ArrayCBORSerializable):
             )
         return ExecutionUnits(self.mem + other.mem, self.steps + other.steps)
 
+    def is_empty(self) -> bool:
+        return self.mem == 0 and self.steps == 0
+
+    def __bool__(self):
+        return not self.is_empty()
+
 
 @dataclass(repr=False)
 class Redeemer(ArrayCBORSerializable):
@@ -751,7 +991,7 @@ class Redeemer(ArrayCBORSerializable):
     ex_units: Optional[ExecutionUnits] = None
 
     @classmethod
-    @limit_primitive_type(list)
+    @limit_primitive_type(list, tuple)
     def from_primitive(cls: Type[Redeemer], values: list) -> Redeemer:
         if isinstance(values[2], CBORTag) and cls is Redeemer:
             values[2] = RawPlutusData.from_primitive(values[2])
@@ -761,13 +1001,62 @@ class Redeemer(ArrayCBORSerializable):
         return redeemer
 
 
-def plutus_script_hash(
-    script: Union[bytes, PlutusV1Script, PlutusV2Script]
-) -> ScriptHash:
+@dataclass(repr=False)
+class RedeemerKey(ArrayCBORSerializable):
+    tag: RedeemerTag
+
+    index: int = field(default=0)
+
+    @classmethod
+    @limit_primitive_type(list, tuple)
+    def from_primitive(cls: Type[RedeemerKey], values: list) -> RedeemerKey:
+        tag = RedeemerTag.from_primitive(values[0])
+        index = values[1]
+        return cls(tag, index)
+
+    def __eq__(self, other):
+        if not isinstance(other, RedeemerKey):
+            return False
+        return self.tag == other.tag and self.index == other.index
+
+    def __hash__(self):
+        return hash(self.to_cbor())
+
+
+@dataclass(repr=False)
+class RedeemerValue(ArrayCBORSerializable):
+    data: Any
+
+    ex_units: ExecutionUnits
+
+    @classmethod
+    @limit_primitive_type(list, tuple)
+    def from_primitive(cls: Type[RedeemerValue], values: list) -> RedeemerValue:
+        if isinstance(values[0], CBORTag) and cls is RedeemerValue:
+            values[0] = RawPlutusData.from_primitive(values[0])
+        return super(RedeemerValue, cls).from_primitive([values[0], values[1]])
+
+    def __eq__(self, other):
+        if not isinstance(other, RedeemerValue):
+            return False
+        return self.data == other.data and self.ex_units == other.ex_units
+
+
+@typechecked
+class RedeemerMap(DictCBORSerializable):
+    KEY_TYPE = RedeemerKey
+
+    VALUE_TYPE = RedeemerValue
+
+
+Redeemers = Union[List[Redeemer], RedeemerMap]
+
+
+def plutus_script_hash(script: Union[NativeScript, PlutusScript]) -> ScriptHash:
     """Calculates the hash of a Plutus script.
 
     Args:
-        script (Union[bytes, PlutusV1Script, PlutusV2Script]): A plutus script.
+        script (Union[bytes, PlutusScript]): A plutus script.
 
     Returns:
         ScriptHash: blake2b hash of the script.
@@ -775,15 +1064,81 @@ def plutus_script_hash(
     return script_hash(script)
 
 
-class PlutusV1Script(bytes):
-    pass
+class PlutusScript(CBORSerializable, bytes):
+    """
+    Plutus script class.
+
+    This class is a base class for all Plutus script versions.
+
+    Example - Load a Plutus script from `test/resources/scriptV2.plutus <https://github.com/Python-Cardano/pycardano/blob/main/test/resources/scriptV2.plutus>`_ and get its address: # noqa: E501
 
 
-class PlutusV2Script(bytes):
-    pass
+        >>> from pycardano import Address, Network
+        >>> script = PlutusV2Script.load("test/resources/scriptV2.plutus")
+        >>> Address(plutus_script_hash(script), network=Network.TESTNET).encode()
+        'addr_test1wrmz3pjz4dmfxj0fc0a0eyw69tp6h7mpndzf9g3kttq9cqqqw47ym'
+    """
+
+    @property
+    def version(self) -> int:
+        raise NotImplementedError("")
+
+    def to_shallow_primitive(self) -> bytes:
+        return bytes(self)
+
+    @classmethod
+    def from_primitive(
+        cls: Type[PlutusScript], value: Any, type_args: Optional[tuple] = None
+    ) -> PlutusScript:
+        if not isinstance(value, (bytes, bytearray)):
+            raise DeserializeException(f"Expect bytes, got {type(value)} instead.")
+        return cls(value)
+
+    @classmethod
+    def from_version(cls, version: int, script_data: bytes) -> "PlutusScript":
+        class_name = f"PlutusV{version}Script"
+        script_class = globals().get(class_name)
+
+        if script_class is None:
+            raise ValueError(f"No Plutus script class found for version {version}")
+
+        return script_class(script_data)
+
+    def get_script_hash_prefix(self) -> bytes:
+        raise NotImplementedError("")
+
+    def __repr__(self):
+        return f"{self.__class__.__name__}({self.hex()})"
 
 
-ScriptType = Union[bytes, NativeScript, PlutusV1Script, PlutusV2Script]
+class PlutusV1Script(PlutusScript):
+    def get_script_hash_prefix(self) -> bytes:
+        return bytes.fromhex("01")
+
+    @property
+    def version(self) -> int:
+        return 1
+
+
+class PlutusV2Script(PlutusScript):
+    def get_script_hash_prefix(self) -> bytes:
+        return bytes.fromhex("02")
+
+    @property
+    def version(self) -> int:
+        return 2
+
+
+class PlutusV3Script(PlutusScript):
+    def get_script_hash_prefix(self) -> bytes:
+        return bytes.fromhex("03")
+
+    @property
+    def version(self) -> int:
+        return 3
+
+
+ScriptType = Union[NativeScript, PlutusScript]
 """Script type. A Union type that contains all valid script types."""
 
 
@@ -798,13 +1153,24 @@ def script_hash(script: ScriptType) -> ScriptHash:
     """
     if isinstance(script, NativeScript):
         return script.hash()
-    elif isinstance(script, PlutusV1Script) or type(script) is bytes:
+    elif isinstance(script, PlutusScript):
+        return ScriptHash(
+            blake2b(
+                script.get_script_hash_prefix() + script,
+                SCRIPT_HASH_SIZE,
+                encoder=RawEncoder,
+            )
+        )
+    elif type(script) is bytes:
         return ScriptHash(
             blake2b(bytes.fromhex("01") + script, SCRIPT_HASH_SIZE, encoder=RawEncoder)
         )
-    elif isinstance(script, PlutusV2Script):
-        return ScriptHash(
-            blake2b(bytes.fromhex("02") + script, SCRIPT_HASH_SIZE, encoder=RawEncoder)
-        )
     else:
         raise TypeError(f"Unexpected script type: {type(script)}")
+
+
+@dataclass
+class Unit(PlutusData):
+    """The default "Unit type" with a 0 constructor ID"""
+
+    CONSTR_ID = 0
