@@ -4,18 +4,18 @@ from __future__ import annotations
 
 from copy import deepcopy
 from dataclasses import dataclass, field
-from pprint import pformat
 from typing import Any, Callable, List, Optional, Type, Union
 
 import cbor2
 from cbor2 import CBORTag
 from nacl.encoding import RawEncoder
 from nacl.hash import blake2b
-from typeguard import typechecked
+from pprintpp import pformat
 
 from pycardano.address import Address
 from pycardano.certificate import Certificate
-from pycardano.exception import InvalidDataException, InvalidOperationException
+from pycardano.exception import InvalidDataException
+from pycardano.governance import ProposalProcedure, VotingProcedures
 from pycardano.hash import (
     TRANSACTION_HASH_SIZE,
     AuxiliaryDataHash,
@@ -29,16 +29,21 @@ from pycardano.hash import (
 from pycardano.metadata import AuxiliaryData
 from pycardano.nativescript import NativeScript
 from pycardano.network import Network
-from pycardano.plutus import Datum, PlutusV1Script, PlutusV2Script, RawPlutusData
+from pycardano.plutus import Datum, PlutusScript, RawPlutusData
 from pycardano.serialization import (
     ArrayCBORSerializable,
     CBORSerializable,
+    DictBase,
     DictCBORSerializable,
     MapCBORSerializable,
+    NonEmptyOrderedSet,
+    OrderedSet,
     Primitive,
     default_encoder,
+    limit_primitive_type,
     list_hook,
 )
+from pycardano.types import typechecked
 from pycardano.witness import TransactionWitnessSet
 
 __all__ = [
@@ -54,6 +59,9 @@ __all__ = [
     "Withdrawals",
 ]
 
+_MAX_INT64 = (1 << 63) - 1
+_MIN_INT64 = -(1 << 63)
+
 
 @dataclass(repr=False)
 class TransactionInput(ArrayCBORSerializable):
@@ -78,6 +86,14 @@ class Asset(DictCBORSerializable):
 
     VALUE_TYPE = int
 
+    def normalize(self) -> Asset:
+        """Normalize the Asset by removing zero values."""
+        for k, v in list(self.items()):
+            if v == 0:
+                self.pop(k)
+
+        return self
+
     def union(self, other: Asset) -> Asset:
         return self + other
 
@@ -85,25 +101,18 @@ class Asset(DictCBORSerializable):
         new_asset = deepcopy(self)
         for n in other:
             new_asset[n] = new_asset.get(n, 0) + other[n]
-        return new_asset
+        return new_asset.normalize()
 
     def __iadd__(self, other: Asset) -> Asset:
         new_item = self + other
-        self.update(new_item)
-        return self
+        self.data = new_item.data
+        return self.normalize()
 
     def __sub__(self, other: Asset) -> Asset:
         new_asset = deepcopy(self)
         for n in other:
-            if n not in new_asset:
-                raise InvalidOperationException(
-                    f"Asset: {new_asset} does not have asset with name: {n}"
-                )
-            # According to ledger rule, the value of an asset could be negative, so we don't check the value here and
-            # will leave the check to users when necessary.
-            # https://github.com/input-output-hk/cardano-ledger/blob/master/eras/alonzo/test-suite/cddl-files/alonzo.cddl#L378
-            new_asset[n] -= other[n]
-        return new_asset
+            new_asset[n] = new_asset.get(n, 0) - other[n]
+        return new_asset.normalize()
 
     def __eq__(self, other):
         if not isinstance(other, Asset):
@@ -122,6 +131,32 @@ class Asset(DictCBORSerializable):
                 return False
         return True
 
+    def __lt__(self, other: Asset):
+        return self <= other and self != other
+
+    def __ge__(self, other: Asset) -> bool:
+        for n in other:
+            if n not in self or self[n] < other[n]:
+                return False
+        return True
+
+    def __gt__(self, other: Asset) -> bool:
+        return self >= other and self != other
+
+    @classmethod
+    @limit_primitive_type(dict)
+    def from_primitive(cls: Type[DictBase], value: dict) -> DictBase:
+        res = super().from_primitive(value)
+        # pop zero values
+        for n, v in list(res.items()):
+            if v == 0:
+                res.pop(n)
+        return res
+
+    def to_shallow_primitive(self) -> dict:
+        x = deepcopy(self).normalize()
+        return super(self.__class__, x).to_shallow_primitive()
+
 
 @typechecked
 class MultiAsset(DictCBORSerializable):
@@ -132,28 +167,30 @@ class MultiAsset(DictCBORSerializable):
     def union(self, other: MultiAsset) -> MultiAsset:
         return self + other
 
+    def normalize(self) -> MultiAsset:
+        """Normalize the MultiAsset by removing zero values."""
+        for k, v in list(self.items()):
+            v.normalize()
+            if len(v) == 0:
+                self.pop(k)
+        return self
+
     def __add__(self, other):
         new_multi_asset = deepcopy(self)
         for p in other:
-            if p not in new_multi_asset:
-                new_multi_asset[p] = Asset()
-            new_multi_asset[p] += other[p]
-        return new_multi_asset
+            new_multi_asset[p] = new_multi_asset.get(p, Asset()) + other[p]
+        return new_multi_asset.normalize()
 
     def __iadd__(self, other):
         new_item = self + other
-        self.update(new_item)
-        return self
+        self.data = new_item.data
+        return self.normalize()
 
     def __sub__(self, other: MultiAsset) -> MultiAsset:
         new_multi_asset = deepcopy(self)
         for p in other:
-            if p not in new_multi_asset:
-                raise InvalidOperationException(
-                    f"MultiAsset: {new_multi_asset} doesn't have policy: {p}"
-                )
-            new_multi_asset[p] -= other[p]
-        return new_multi_asset
+            new_multi_asset[p] = new_multi_asset.get(p, Asset()) - other[p]
+        return new_multi_asset.normalize()
 
     def __eq__(self, other):
         if not isinstance(other, MultiAsset):
@@ -166,12 +203,28 @@ class MultiAsset(DictCBORSerializable):
                     return False
             return True
 
-    def __le__(self, other: MultiAsset):
-        for p in self:
-            if p not in other or not self[p] <= other[p]:
+    def __ge__(self, other: MultiAsset) -> bool:
+        for n in other:
+            if n not in self:
+                return False
+            if not self[n] >= other[n]:
                 return False
         return True
 
+    def __gt__(self, other: MultiAsset) -> bool:
+        return self >= other and self != other
+
+    def __le__(self, other: MultiAsset):
+        for p in self:
+            if p not in other:
+                return False
+            if not self[p] <= other[p]:
+                return False
+        return True
+
+    def __lt__(self, other: MultiAsset):
+        return self <= other and self != other
+
     def filter(
         self, criteria=Callable[[ScriptHash, AssetName, int], bool]
     ) -> MultiAsset:
@@ -213,6 +266,20 @@ class MultiAsset(DictCBORSerializable):
 
         return count
 
+    @classmethod
+    @limit_primitive_type(dict)
+    def from_primitive(cls: Type[DictBase], value: dict) -> DictBase:
+        res = super().from_primitive(value)
+        # pop empty values
+        for n, v in list(res.items()):
+            if not v:
+                res.pop(n)
+        return res
+
+    def to_shallow_primitive(self) -> dict:
+        x = deepcopy(self).normalize()
+        return super(self.__class__, x).to_shallow_primitive()
+
 
 @typechecked
 @dataclass(repr=False)
@@ -258,6 +325,14 @@ class Value(ArrayCBORSerializable):
     def __lt__(self, other: Union[Value, int]):
         return self <= other and self != other
 
+    def __ge__(self, other: Union[Value, int]):
+        if isinstance(other, int):
+            other = Value(other)
+        return self.coin >= other.coin and self.multi_asset >= other.multi_asset
+
+    def __gt__(self, other: Union[Value, int]):
+        return self >= other and self != other
+
     def to_shallow_primitive(self):
         if self.multi_asset:
             return super().to_shallow_primitive()
@@ -269,25 +344,23 @@ class Value(ArrayCBORSerializable):
 class _Script(ArrayCBORSerializable):
     _TYPE: int = field(init=False, default=0)
 
-    script: Union[NativeScript, PlutusV1Script, PlutusV2Script]
+    script: Union[NativeScript, PlutusScript]
 
     def __post_init__(self):
         if isinstance(self.script, NativeScript):
             self._TYPE = 0
-        elif isinstance(self.script, PlutusV1Script):
-            self._TYPE = 1
-        else:
-            self._TYPE = 2
+        elif isinstance(self.script, PlutusScript):
+            self._TYPE = self.script.version
 
     @classmethod
-    def from_primitive(cls: Type[_Script], values: List[Primitive]) -> _Script:
+    def from_primitive(
+        cls: Type[_Script], values: List[Primitive], type_args: Optional[tuple] = None
+    ) -> _Script:
         if values[0] == 0:
             return cls(NativeScript.from_primitive(values[1]))
         assert isinstance(values[1], bytes)
-        if values[0] == 1:
-            return cls(PlutusV1Script(values[1]))
-        else:
-            return cls(PlutusV2Script(values[1]))
+        assert isinstance(values[0], int)
+        return cls(PlutusScript.from_version(values[0], values[1]))
 
 
 @dataclass(repr=False)
@@ -302,7 +375,8 @@ class _DatumOption(ArrayCBORSerializable):
         else:
             self._TYPE = 1
 
-    def to_shallow_primitive(self) -> List[Primitive]:
+    def to_shallow_primitive(self) -> Primitive:
+        data: Union[CBORTag, DatumHash]
         if self._TYPE == 1:
             data = CBORTag(24, cbor2.dumps(self.datum, default=default_encoder))
         else:
@@ -311,7 +385,9 @@ class _DatumOption(ArrayCBORSerializable):
 
     @classmethod
     def from_primitive(
-        cls: Type[_DatumOption], values: List[Primitive]
+        cls: Type[_DatumOption],
+        values: List[Primitive],
+        type_args: Optional[tuple] = None,
     ) -> _DatumOption:
         if values[0] == 0:
             assert isinstance(values[1], bytes)
@@ -333,7 +409,9 @@ class _ScriptRef(CBORSerializable):
         return CBORTag(24, cbor2.dumps(self.script, default=default_encoder))
 
     @classmethod
-    def from_primitive(cls: Type[_ScriptRef], value: Primitive) -> _ScriptRef:
+    def from_primitive(
+        cls: Type[_ScriptRef], value: List[Primitive], type_args: Optional[tuple] = None
+    ) -> _ScriptRef:
         assert isinstance(value, CBORTag)
         return cls(_Script.from_primitive(cbor2.loads(value.value)))
 
@@ -353,7 +431,9 @@ class _TransactionOutputPostAlonzo(MapCBORSerializable):
     )
 
     @property
-    def script(self) -> Optional[Union[NativeScript, PlutusV1Script, PlutusV2Script]]:
+    def script(
+        self,
+    ) -> Optional[Union[NativeScript, PlutusScript]]:
         if self.script_ref:
             return self.script_ref.script.script
         else:
@@ -379,7 +459,7 @@ class TransactionOutput(CBORSerializable):
 
     datum: Optional[Datum] = None
 
-    script: Optional[Union[NativeScript, PlutusV1Script, PlutusV2Script]] = None
+    script: Optional[Union[NativeScript, PlutusScript]] = None
 
     post_alonzo: Optional[bool] = False
 
@@ -390,11 +470,7 @@ class TransactionOutput(CBORSerializable):
             self.amount = Value(self.amount)
 
     def validate(self):
-        if isinstance(self.amount, int) and self.amount < 0:
-            raise InvalidDataException(
-                f"Transaction output cannot have negative amount of ADA or "
-                f"native asset: \n {self.amount}"
-            )
+        super().validate()
         if isinstance(self.amount, Value) and (
             self.amount.coin < 0
             or self.amount.multi_asset.count(lambda p, n, v: v < 0) > 0
@@ -406,10 +482,7 @@ class TransactionOutput(CBORSerializable):
 
     @property
     def lovelace(self) -> int:
-        if isinstance(self.amount, int):
-            return self.amount
-        else:
-            return self.amount.coin
+        return self.amount.coin
 
     def to_primitive(self) -> Primitive:
         if self.datum or self.script or self.post_alonzo:
@@ -431,11 +504,13 @@ class TransactionOutput(CBORSerializable):
 
     @classmethod
     def from_primitive(
-        cls: Type[TransactionOutput], value: Primitive
+        cls: Type[TransactionOutput],
+        value: List[Primitive],
+        type_args: Optional[tuple] = None,
     ) -> TransactionOutput:
         if isinstance(value, list):
             output = _TransactionOutputLegacy.from_primitive(value)
-            return cls(output.address, output.amount, datum=output.datum_hash)
+            return cls(output.address, output.amount, datum_hash=output.datum_hash)
         else:
             output = _TransactionOutputPostAlonzo.from_primitive(value)
             datum = output.datum.datum if output.datum else None
@@ -465,9 +540,7 @@ class UTxO(ArrayCBORSerializable):
         return pformat(vars(self))
 
     def __hash__(self):
-        return hash(
-            blake2b(self.input.to_cbor("bytes") + self.output.to_cbor("bytes"), 32)
-        )
+        return hash(blake2b(self.input.to_cbor() + self.output.to_cbor(), 32))
 
 
 class Withdrawals(DictCBORSerializable):
@@ -490,9 +563,9 @@ class Withdrawals(DictCBORSerializable):
 
 @dataclass(repr=False)
 class TransactionBody(MapCBORSerializable):
-    inputs: List[TransactionInput] = field(
-        default_factory=list,
-        metadata={"key": 0, "object_hook": list_hook(TransactionInput)},
+    inputs: Union[List[TransactionInput], OrderedSet[TransactionInput]] = field(
+        default_factory=OrderedSet,
+        metadata={"key": 0},
     )
 
     outputs: List[TransactionOutput] = field(
@@ -504,15 +577,20 @@ class TransactionBody(MapCBORSerializable):
 
     ttl: Optional[int] = field(default=None, metadata={"key": 3, "optional": True})
 
-    certificates: Optional[List[Certificate]] = field(
-        default=None, metadata={"key": 4, "optional": True}
+    certificates: Optional[
+        Union[List[Certificate], NonEmptyOrderedSet[Certificate]]
+    ] = field(
+        default=None,
+        metadata={
+            "key": 4,
+            "optional": True,
+        },
     )
 
     withdraws: Optional[Withdrawals] = field(
         default=None, metadata={"key": 5, "optional": True}
     )
 
-    # TODO: Add proposal update support
     update: Any = field(default=None, metadata={"key": 6, "optional": True})
 
     auxiliary_data_hash: Optional[AuxiliaryDataHash] = field(
@@ -531,21 +609,23 @@ class TransactionBody(MapCBORSerializable):
         default=None, metadata={"key": 11, "optional": True}
     )
 
-    collateral: Optional[List[TransactionInput]] = field(
+    collateral: Optional[
+        Union[List[TransactionInput], NonEmptyOrderedSet[TransactionInput]]
+    ] = field(
         default=None,
         metadata={
             "key": 13,
             "optional": True,
-            "object_hook": list_hook(TransactionInput),
         },
     )
 
-    required_signers: Optional[List[VerificationKeyHash]] = field(
+    required_signers: Optional[
+        Union[List[VerificationKeyHash], NonEmptyOrderedSet[VerificationKeyHash]]
+    ] = field(
         default=None,
         metadata={
             "key": 14,
             "optional": True,
-            "object_hook": list_hook(VerificationKeyHash),
         },
     )
 
@@ -561,17 +641,43 @@ class TransactionBody(MapCBORSerializable):
         default=None, metadata={"key": 17, "optional": True}
     )
 
-    reference_inputs: Optional[List[TransactionInput]] = field(
+    reference_inputs: Optional[
+        Union[List[TransactionInput], NonEmptyOrderedSet[TransactionInput]]
+    ] = field(
         default=None,
         metadata={
             "key": 18,
-            "object_hook": list_hook(TransactionInput),
             "optional": True,
         },
     )
 
+    voting_procedures: Optional[VotingProcedures] = field(
+        default=None, metadata={"key": 19, "optional": True}
+    )
+
+    proposal_procedures: Optional[NonEmptyOrderedSet[ProposalProcedure]] = field(
+        default=None, metadata={"key": 20, "optional": True}
+    )
+
+    current_treasury_value: Optional[int] = field(
+        default=None, metadata={"key": 21, "optional": True}
+    )
+
+    donation: Optional[int] = field(
+        default=None, metadata={"key": 22, "optional": True}
+    )
+
+    def validate(self):
+        if (
+            self.mint
+            and self.mint.count(lambda p, n, v: v < _MIN_INT64 or v > _MAX_INT64) > 0
+        ):
+            raise InvalidDataException(
+                f"Mint amount must be between {_MIN_INT64} and {_MAX_INT64}. \n Mint amount: {self.mint}"
+            )
+
     def hash(self) -> bytes:
-        return blake2b(self.to_cbor(encoding="bytes"), TRANSACTION_HASH_SIZE, encoder=RawEncoder)  # type: ignore
+        return blake2b(self.to_cbor(), TRANSACTION_HASH_SIZE, encoder=RawEncoder)  # type: ignore
 
     @property
     def id(self) -> TransactionId:
@@ -584,10 +690,22 @@ class Transaction(ArrayCBORSerializable):
 
     transaction_witness_set: TransactionWitnessSet
 
-    valid: bool = True
+    valid: Optional[bool] = field(default=True, metadata={"optional": True})
 
     auxiliary_data: Optional[AuxiliaryData] = None
 
+    @property
+    def json_type(self) -> str:
+        return (
+            "Unwitnessed Tx ConwayEra"
+            if self.transaction_witness_set.vkey_witnesses is None
+            else "Signed Tx ConwayEra"
+        )
+
+    @property
+    def json_description(self) -> str:
+        return "Ledger Cddl Format"
+
     @property
     def id(self) -> TransactionId:
         return self.transaction_body.id
